/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Netscape Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s):
 *   Peter Hartshorn <peter@igelaus.com.au>
 *   B.J. Rossiter <bj@igelaus.com.au>
 *   Matthew Wlazlo <mattw@igelaus.com.au>
 */

#include "nsRenderingContextNanoX.h"
#include "nsDrawingSurfaceNanoX.h"
#include "nsDeviceContextNanoX.h"
#include "nsIPref.h"
#include "nsIServiceManager.h"
#include "nsGfxCIID.h"
#include "nspr.h"
#include "nsCRT.h"

#include <nano-X.h>
#include <stdlib.h>
#include <stdio.h>

#include "nsGfxPSCID.h"
#include "nsIDeviceContextPS.h"

nscoord nsDeviceContextNanoX::mDpi = 96;

static NS_DEFINE_CID(kPrefCID, NS_PREF_CID);
static NS_DEFINE_IID(kDeviceContextIID, NS_IDEVICE_CONTEXT_IID);

NS_IMPL_ISUPPORTS1(nsDeviceContextNanoX, nsIDeviceContext)

nsDeviceContextNanoX::nsDeviceContextNanoX() //: DeviceContextImpl()
{
  NS_INIT_REFCNT();
  mTwipsToPixels = 1.0;
  mPixelsToTwips = 1.0;
  mDepth = 0;
  mPaletteInfo.isPaletteDevice = PR_FALSE;
  mPaletteInfo.sizePalette = 0;
  mPaletteInfo.numReserved = 0;
  mPaletteInfo.palette = nsnull;
  mNumCells = 0;

  mWidthFloat = 0.0f;
  mHeightFloat = 0.0f;
  mWidth = -1;
  mHeight = -1;
  mDeviceWindow = 0;
}

nsDeviceContextNanoX::~nsDeviceContextNanoX()
{
  //nsDrawingSurfaceNanoX *surf = (nsDrawingSurfaceNanoX *)mSurface;
  //NS_IF_RELEASE(surf);
  //mSurface = nsnull;
  nsresult rv;
  nsCOMPtr<nsIPref> prefs = do_GetService(kPrefCID, &rv);
  if (NS_SUCCEEDED(rv)) {
    prefs->UnregisterCallback("browser.display.screen_resolution",
                              prefChanged, (void *)this);
  }
}

NS_IMETHODIMP nsDeviceContextNanoX::Init(nsNativeWidget aNativeWidget)
{
  // get the screen object and its width/height
  // XXXRight now this will only get the primary monitor.

  if (!mScreenManager)
    mScreenManager = do_GetService("@mozilla.org/gfx/screenmanager;1");
  if (!mScreenManager) {
    return NS_ERROR_FAILURE;
  }

  if (aNativeWidget) {
    mDeviceWindow = (GR_WINDOW_ID)aNativeWidget;
  }

  nsCOMPtr<nsIScreen> screen;
  mScreenManager->GetPrimaryScreen ( getter_AddRefs(screen) );
  if ( screen ) {
    PRInt32 x, y, width, height, depth;
    screen->GetAvailRect ( &x, &y, &width, &height );
    screen->GetPixelDepth ( &depth );
    mWidthFloat = float(width);
    mHeightFloat = float(height);
    mDepth = NS_STATIC_CAST ( PRUint32, depth );
  }
    
  static int initialized = 0;
  if (!initialized) {
    initialized = 1;

    // Set prefVal the value of the preference
    // "browser.display.screen_resolution"
    // or -1 if we can't get it.
    // If it's negative, we pretend it's not set.
    // If it's 0, it means force use of the operating system's logical
    // resolution.
    // If it's positive, we use it as the logical resolution
    PRInt32 prefVal = -1;
    nsresult res;

    nsCOMPtr<nsIPref> prefs(do_GetService(kPrefCID, &res));
    if (NS_SUCCEEDED(res) && prefs) {
      res = prefs->GetIntPref("browser.display.screen_resolution", &prefVal);
      if (! NS_SUCCEEDED(res)) {
        prefVal = -1;
      }
      prefs->RegisterCallback("browser.display.screen_resolution", prefChanged,
                              (void *)this);
    }
    // XXX This is broken (obvously), needs fixing.
#if 0
    // Set OSVal to what the operating system thinks the logical resolution is.
    float screenWidthIn; 
    {
    	GR_SCREEN_INFO si;
	GrGetScreenInfo(&si);
	float sw_mm = si.xdpcm / 10;
	screenWidthIn = sw_mm / 25.4f;
    }
    //float screenWidthIn = float(::gdk_screen_width_mm()) / 25.4f;
    PRInt32 OSVal = nscoord(mWidthFloat / screenWidthIn);

    if (prefVal > 0) {
      // If there's a valid pref value for the logical resolution,
      // use it.
      mDpi = prefVal;
    } else if ((prefVal == 0) || (OSVal > 96)) {
      // Either if the pref is 0 (force use of OS value) or the OS
      // value is bigger than 96, use the OS value.
      mDpi = OSVal;
    } else {
      // if we couldn't get the pref or it's negative, and the OS
      // value is under 96ppi, then use 96.
#endif
      mDpi = 96;
//    }
  }

  SetDPI(mDpi);
  
  /*
  sb = gtk_vscrollbar_new(NULL);
  gtk_widget_ref(sb);
  gtk_object_sink(GTK_OBJECT(sb));
  gtk_widget_size_request(sb,&req);
  mScrollbarWidth = req.width;
  gtk_widget_destroy(sb);
  gtk_widget_unref(sb);
  
  sb = gtk_hscrollbar_new(NULL);
  gtk_widget_ref(sb);
  gtk_object_sink(GTK_OBJECT(sb));
  gtk_widget_size_request(sb,&req);
  mScrollbarHeight = req.height;
  gtk_widget_destroy(sb);
  gtk_widget_unref(sb);
  */
  mScrollbarHeight = 10;
  mScrollbarWidth = 10;
#ifdef DEBUG
  static PRBool once = PR_TRUE;
  if (once) {
    printf("GFX: dpi=%d t2p=%g p2t=%g depth=%d\n", mDpi, mTwipsToPixels, mPixelsToTwips,mDepth);
    once = PR_FALSE;
  }
#endif

  DeviceContextImpl::CommonInit();

  return NS_OK;
}


NS_IMETHODIMP nsDeviceContextNanoX::CreateRenderingContext(nsIRenderingContext *&aContext)
{
  nsIRenderingContext *context = nsnull;
  nsDrawingSurfaceNanoX *surface = nsnull;
  nsresult rv;

  GR_WINDOW_ID wid = (GR_WINDOW_ID)mWidget;

  context = new nsRenderingContextNanoX();

  if (context) {
    NS_ADDREF(context);
    surface = new nsDrawingSurfaceNanoX();
    if (surface && wid) {
      rv = surface->Init((GR_WINDOW_ID)mWidget);

      if (NS_OK == rv) {
        rv = context->Init(this, surface);
      }
    } else {
     rv = NS_ERROR_OUT_OF_MEMORY;
    }
  } else {
    rv = NS_ERROR_OUT_OF_MEMORY;
  }

  if(NS_OK != rv)
  	NS_IF_RELEASE(context);

  aContext = context;
  return rv;
}

NS_IMETHODIMP nsDeviceContextNanoX::SupportsNativeWidgets(PRBool &aSupportsWidgets)
{
  aSupportsWidgets = PR_TRUE;
  return NS_OK;
}

NS_IMETHODIMP nsDeviceContextNanoX::GetScrollBarDimensions(float &aWidth,
                                                           float &aHeight) const
{
//  aWidth = 15 * mPixelsToTwips;
//  aHeight = 15 * mPixelsToTwips;
  aWidth = mScrollbarWidth * mPixelsToTwips;
  aHeight = mScrollbarHeight * mPixelsToTwips;
  return NS_OK;
}

NS_IMETHODIMP nsDeviceContextNanoX::GetSystemAttribute(nsSystemAttrID anID,
                                                 SystemAttrStruct *aInfo) const
{
  nsresult status = NS_OK;

  switch (anID) {
  case eSystemAttr_Color_WindowBackground:
        *aInfo->mColor = NS_RGB(255,255,255);
        break;
    case eSystemAttr_Color_WindowForeground:
        *aInfo->mColor = NS_RGB(0,0,0);
        break;
    case eSystemAttr_Color_WidgetBackground:
      *aInfo->mColor = NS_RGB(255,255,255);
        break;
    case eSystemAttr_Color_WidgetForeground:
        *aInfo->mColor = NS_RGB(0,0,0);
        break;
    case eSystemAttr_Color_WidgetSelectBackground:
      *aInfo->mColor = NS_RGB(255,255,255);
        break;
    case eSystemAttr_Color_WidgetSelectForeground:
        *aInfo->mColor = NS_RGB(0,0,0);
        break;
    case eSystemAttr_Color_Widget3DHighlight:
        *aInfo->mColor = NS_RGB(0xa0,0xa0,0xa0);
        break;
    case eSystemAttr_Color_Widget3DShadow:
        *aInfo->mColor = NS_RGB(0x40,0x40,0x40);
        break;
    case eSystemAttr_Color_TextBackground:
      *aInfo->mColor = NS_RGB(255,255,255);
        break;
    case eSystemAttr_Color_TextForeground:
        *aInfo->mColor = NS_RGB(0,0,0);
        break;
    case eSystemAttr_Color_TextSelectBackground:
      *aInfo->mColor = NS_RGB(255,255,255);
        break;
    case eSystemAttr_Color_TextSelectForeground:
        *aInfo->mColor = NS_RGB(0,0,0);
        break;
    //---------
    // Size
    //---------
    case eSystemAttr_Size_ScrollbarHeight:
        aInfo->mSize = mScrollbarHeight;
        break;
    case eSystemAttr_Size_ScrollbarWidth:
        aInfo->mSize = mScrollbarWidth;
        break;
    case eSystemAttr_Size_WindowTitleHeight:
        aInfo->mSize = 0;
        break;
    case eSystemAttr_Size_WindowBorderWidth:
      //      aInfo->mSize = style->klass->xthickness;
      aInfo->mSize = 1;
        break;
    case eSystemAttr_Size_WindowBorderHeight:
      //        aInfo->mSize = style->klass->ythickness;
        aInfo->mSize = 1;
        break;
    case eSystemAttr_Size_Widget3DBorder:
        aInfo->mSize = 4;
        break;
    //---------
    // Fonts
    //---------
    case eSystemAttr_Font_Caption: // css2
    case eSystemAttr_Font_Icon:
    case eSystemAttr_Font_Menu:
    case eSystemAttr_Font_MessageBox:
    case eSystemAttr_Font_SmallCaption:
    case eSystemAttr_Font_StatusBar:

    case eSystemAttr_Font_Window:			// css3
    case eSystemAttr_Font_Document:
    case eSystemAttr_Font_Workspace:
    case eSystemAttr_Font_Desktop:
    case eSystemAttr_Font_Info:
    case eSystemAttr_Font_Dialog:
    case eSystemAttr_Font_Button:
    case eSystemAttr_Font_PullDownMenu:
    case eSystemAttr_Font_List:
    case eSystemAttr_Font_Field:

    case eSystemAttr_Font_Tooltips: // moz
    case eSystemAttr_Font_Widget:
#if 1
	return NS_OK;
#else
	    status = NS_ERROR_FAILURE;
	    nsresult res;
	    NS_WITH_SERVICE(nsIPref, prefs, kPrefCID, &res);
	    if (NS_SUCCEEDED(res) && prefs) {
	      char *cVal = 0;// = (char*)malloc(100*sizeof(char));
	      res = prefs->CopyCharPref("font.name.monospace.x-western", &cVal);
	      if (NS_SUCCEEDED(res)) {
		if (cVal) {
		  // check to see if we have the font..
		  GR_FONT_ID fid = 0;
		  fid = GrCreateFont((GR_CHAR*)cVal, 0, NULL);
		  GrDestroyFont(fid);
		  free(cVal);
		  cVal = 0;
		  status = (fid == 0 ? NS_ERROR_FAILURE : NS_OK);
		} 
	      }
	    }
#endif
      break;
  } // switch

  return status;
}

NS_IMETHODIMP nsDeviceContextNanoX::GetDrawingSurface(nsIRenderingContext &aContext, nsDrawingSurface &aSurface)
{
  aContext.CreateDrawingSurface(nsnull, 0, aSurface);
  return nsnull == aSurface ? NS_ERROR_OUT_OF_MEMORY : NS_OK;  
/*
  if (!mSurface) {
    aContext.CreateDrawingSurface(nsnull, 0, mSurface);
  }
  aSurface = mSurface;
  return NS_OK;
*/
}

NS_IMETHODIMP nsDeviceContextNanoX::ConvertPixel(nscolor aColor,
                                                 PRUint32 &aPixel)
{
  aPixel = MWRGB(NS_GET_R(aPixel),
                 NS_GET_G(aPixel),
                 NS_GET_B(aPixel));
  return NS_OK;
}

NS_IMETHODIMP nsDeviceContextNanoX::CheckFontExistence(const nsString& aFontName)
{
  char *fontName = aFontName.ToNewCString();
  GR_FONT_ID fid = 0;

  fid = GrCreateFont((GR_CHAR*)fontName, 0, NULL);

  GrDestroyFont(fid);
  
  return fid == 0 ? NS_ERROR_FAILURE : NS_OK;
}

NS_IMETHODIMP nsDeviceContextNanoX::GetDeviceSurfaceDimensions(PRInt32 &aWidth,
                                                               PRInt32 &aHeight)
{
  if (mWidth == -1)
    mWidth = NSToIntRound(mWidthFloat * mDevUnitsToAppUnits);

  if (mHeight == -1)
    mHeight = NSToIntRound(mHeightFloat * mDevUnitsToAppUnits);

  aWidth = mWidth;
  aHeight = mHeight;

  return NS_OK;
}

NS_IMETHODIMP nsDeviceContextNanoX::GetRect(nsRect &aRect)
{
  // if we have an initialized widget for this device context, use it
  // to try and get real screen coordinates.
  if(mDeviceWindow) {
    int x, y, width, height, depth;
    x = y = width = height = 0;

    {
	GR_WINDOW_INFO info;
	GrGetWindowInfo(mDeviceWindow, &info);
	x = info.x;
	y = info.y;
	width = info.width;
	height = info.height;
    }

    {
    	GR_SCREEN_INFO info;
	GrGetScreenInfo(&info);
	depth = info.planes;
    }

    
    //gdk_window_get_geometry(mDeviceWindow, &x, &y, &width, &height,
    //                        &depth);
    //gdk_window_get_origin(mDeviceWindow, &x, &y);

    nsCOMPtr<nsIScreen> screen;
    mScreenManager->ScreenForRect(x, y, width, height, getter_AddRefs(screen));
    screen->GetRect(&aRect.x, &aRect.y, &aRect.width, &aRect.height);
    aRect.x = NSToIntRound(mDevUnitsToAppUnits * aRect.x);
    aRect.y = NSToIntRound(mDevUnitsToAppUnits * aRect.y);
    aRect.width = NSToIntRound(mDevUnitsToAppUnits * aRect.width);
    aRect.height = NSToIntRound(mDevUnitsToAppUnits * aRect.height);

  } else {
	  PRInt32 width, height;
	  GetDeviceSurfaceDimensions(width, height);
	  aRect.x = 0;
	  aRect.y = 0;
	  aRect.width = width;
	  aRect.height = height;
  }
  return NS_OK;
}

NS_IMETHODIMP nsDeviceContextNanoX::GetClientRect(nsRect &aRect)
{
  // TODO: Is this true for NanoX? I would assume so.. but check anyway. MattW
  // The client rect is never different from the standard rect on
  // linux because we don't have the concept of the title bar.
  return GetRect(aRect);
}

NS_IMETHODIMP nsDeviceContextNanoX::GetDeviceContextFor(nsIDeviceContextSpec *aDevice, nsIDeviceContext *&aContext)
{
#ifdef USE_XPRINT
  int method=-1;
  nsDeviceContextSpecNanoX *spec = NS_STATIC_CAST(nsDeviceContextSpecNanoX *, aDevice);
  spec->GetPrintMethod(method);

  if (method == 1) { // XPRINT
    static NS_DEFINE_CID(kCDeviceContextXp, NS_DEVICECONTEXTXP_CID);
    nsresult rv;
    nsCOMPtr<nsIDeviceContextXp> dcxp(do_CreateInstance(kCDeviceContextXp, &rv));

    NS_ASSERTION(NS_SUCCEEDED(rv), "Couldn't create Xp Device context");    
    if (NS_FAILED(rv)) 
      return rv;

    dcxp->SetSpec(aDevice);
    dcxp->InitDeviceContextXP((nsIDeviceContext*)aContext,
                                (nsIDeviceContext*)this);
  
    rv = dcxp->QueryInterface(NS_GET_IID(nsIDeviceContext),
                              (void **)&aContext);
    return rv;
  }
  // default/PS
#endif /* USE_XPRINT */
  static NS_DEFINE_CID(kCDeviceContextPS, NS_DEVICECONTEXTPS_CID);

  // Create a Postscript device context 
  nsresult rv;
  nsIDeviceContextPS *dcps;

  rv = nsComponentManager::CreateInstance(kCDeviceContextPS,
                                          nsnull,
                                          NS_GET_IID(nsIDeviceContextPS),
                                          (void**)&dcps);

  NS_ASSERTION(NS_SUCCEEDED(rv), "Couldn't create PS Device Context");

  dcps->SetSpec(aDevice);
  dcps->InitDeviceContextPS((nsIDeviceContext*)aContext,
                            (nsIDeviceContext*)this);

  rv = dcps->QueryInterface(NS_GET_IID(nsIDeviceContext),
                            (void **)&aContext);

  NS_RELEASE(dcps);

  return rv;
}

NS_IMETHODIMP nsDeviceContextNanoX::BeginDocument(PRUnichar *aTitle)
{
  return NS_OK;
}

NS_IMETHODIMP nsDeviceContextNanoX::EndDocument(void)
{
  return NS_OK;
}

NS_IMETHODIMP nsDeviceContextNanoX::BeginPage(void)
{
  return NS_OK;
}

NS_IMETHODIMP nsDeviceContextNanoX::EndPage(void)
{
  return NS_OK;
}

NS_IMETHODIMP nsDeviceContextNanoX::GetDepth(PRUint32& aDepth)
{
  aDepth = mDepth;
  return NS_OK;
}

nsresult
nsDeviceContextNanoX::SetDPI(PRInt32 aDpi)
{
  mDpi = aDpi;
  
  int pt2t = 72;

  // make p2t a nice round number - this prevents rounding problems
  mPixelsToTwips = float(NSToIntRound(float(NSIntPointsToTwips(pt2t)) / float(aDpi)));
  mTwipsToPixels = 1.0f / mPixelsToTwips;

  // XXX need to reflow all documents
  return NS_OK;
}

int nsDeviceContextNanoX::prefChanged(const char *aPref, void *aClosure)
{
  nsDeviceContextNanoX *context = (nsDeviceContextNanoX*)aClosure;
  nsresult rv;
  
  if (nsCRT::strcmp(aPref, "browser.display.screen_resolution")==0) {
    PRInt32 dpi;
    nsCOMPtr<nsIPref> prefs(do_GetService(kPrefCID, &rv));
    rv = prefs->GetIntPref(aPref, &dpi);
    if (NS_SUCCEEDED(rv))
      context->SetDPI(dpi);
  }
  
  return 0;
}
