/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Netscape Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s):
 *   Peter Hartshorn <peter@igelaus.com.au>
 *   B.J. Rossiter <bj@igelaus.com.au>
 *   Matthew Wlazlo <mattw@igelaus.com.au>
 */

#include "nsImageNanoX.h"
#include "nsDrawingSurfaceNanoX.h"
#include "nsMemory.h"
#include <stdlib.h>
#include <device.h>

#undef DEBUG_mattw
#undef ANNOYING_PRINTFS

#ifdef DEBUG_mattw
#	define ENTER {fprintf(stderr, "Entering: %s\n", __PRETTY_FUNCTION__); }
#	define LEAVE {fprintf(stderr, "Leaving: %s\n", __PRETTY_FUNCTION__); }
#else
#	undef ANNOYING_PRINTFS
#	define ENTER
#	define LEAVE
#endif


#define IsFlagSet(a,b) ((a) & (b))

static const int gPixelSize = sizeof(GR_PIXELVAL); 

extern
void RectStretch(long xs1,long ys1,long xs2,long ys2,
                 long xd1,long yd1,long xd2,long yd2,
                 unsigned char *aSrcImage, unsigned aSrcStride,
                 unsigned char *aDstImage, unsigned aDstStride,
                 unsigned aDepth);

NS_IMPL_ISUPPORTS1(nsImageNanoX, nsIImage);


nsImageNanoX::nsImageNanoX()
{
	NS_INIT_REFCNT();
	mImageBits = nsnull;
	mAlphaBits = nsnull;
	mWidth = 0;
	mHeight = 0;
	mDepth = 0;
	mRowBytes = 0;
	mSizeImage = 0;
	mNumBytesPixel = 0;
	mImagePixmap = 0;
	mAlphaRegion = 0;
	mAlphaDepth = 0;
	mAlphaRowBytes = 0;
	mAlphaWidth = 0;
	mAlphaHeight = 0;
	mLocation.x = 0;
	mLocation.y = 0;
	mConvertedBits = nsnull;
	mNaturalWidth = 0;
	mNaturalHeight = 0;
	mAlphaValid = PR_FALSE;
	mIsSpacer = PR_TRUE;
	mGC = 0;
}

nsImageNanoX::~nsImageNanoX()
{
	if (nsnull != mImageBits) {
		delete[] mImageBits;
		mImageBits = nsnull;
	}
	if (nsnull != mAlphaBits) {
		delete[] mAlphaBits;
		mAlphaBits = nsnull;

		if (mAlphaRegion != 0) {
			GrDestroyRegion(mAlphaRegion);
			mAlphaRegion = 0;
		}
	}

	if (mImagePixmap != 0) {
		GrDestroyWindow(mImagePixmap);
		mImagePixmap = 0;
	}

	if (mGC) {
		GrDestroyGC(mGC);
	}
}

NS_IMETHODIMP nsImageNanoX::Init(PRInt32 aWidth, PRInt32 aHeight,
                                 PRInt32 aDepth,
                                 nsMaskRequirements aMaskRequirements)
{
	if (nsnull != mImageBits) {
		delete[] mImageBits;
		mImageBits = nsnull;
	}
	if (nsnull != mAlphaBits) {
		delete[] mAlphaBits;
		mAlphaBits = nsnull;
	}
	if (0 != mAlphaRegion) {
		GrDestroyRegion(mAlphaRegion);
		mAlphaRegion = 0;
	}

	SetDecodedRect(0, 0, 0, 0);

	if (0 != mImagePixmap) {
		GrDestroyWindow(mImagePixmap);
		mImagePixmap = 0;
	}

	GR_SCREEN_INFO si;
	GrGetScreenInfo(&si);
	mPixType = si.pixtype;
	mNumBytesPixel = si.bpp / 8;
#ifdef DEBUG_mattw
	fprintf(stderr, "Using pixelformat: %d\n", mPixType);
#endif

	mWidth = aWidth;
	mHeight = aHeight;
	mDepth = aDepth;
	mIsTopToBottom = PR_TRUE;

	ComputeMetrics();

	mImageBits = (PRUint8*)new PRUint8[mSizeImage];

	switch (aMaskRequirements) {
		case nsMaskRequirements_kNoMask:
			mAlphaBits = nsnull;
			mAlphaWidth = 0;
			mAlphaHeight = 0;
			break;
		case nsMaskRequirements_kNeeds1Bit:
			mAlphaRowBytes = (aWidth + 7) / 8;
			mAlphaDepth = 1;

			mAlphaRowBytes = (mAlphaRowBytes + 3) & ~0x3;
			mAlphaBits = new unsigned char[mAlphaRowBytes * aHeight];
			mAlphaWidth = aWidth;
			mAlphaHeight = aHeight;
			break;
		case nsMaskRequirements_kNeeds8Bit:
			mAlphaRowBytes = aWidth;
			mAlphaDepth = 8;
			
			// XXX
			//mAlphaRowBytes = (mAlphaRowBytes + 3) & ~0x3;
			mAlphaBits = new unsigned char[mAlphaRowBytes * aHeight];
			mAlphaWidth = aWidth;
			mAlphaHeight = aHeight;
			break;
	}
	return NS_OK;
}

PRInt32 nsImageNanoX::GetHeight()
{
	return mHeight;
}

PRInt32 nsImageNanoX::GetWidth()
{
	return mWidth;
}

PRUint8 *nsImageNanoX::GetBits()
{
	return mImageBits;
}

void *nsImageNanoX::GetBitInfo()
{
	return nsnull;
}

PRInt32 nsImageNanoX::GetLineStride()
{
	return mRowBytes;
}

nsColorMap *nsImageNanoX::GetColorMap()
{ 
	return nsnull;
}

PRBool nsImageNanoX::IsOptimized()
{
	return PR_TRUE;
}

PRUint8 *nsImageNanoX::GetAlphaBits()
{
	return mAlphaBits;
}

PRInt32 nsImageNanoX::GetAlphaWidth()
{
	return mAlphaWidth;
}

PRInt32 nsImageNanoX::GetAlphaHeight()
{
	return mAlphaHeight;
}

PRInt32 nsImageNanoX::GetAlphaLineStride()
{
	return mAlphaRowBytes;
}

nsIImage *nsImageNanoX::DuplicateImage()
{
	return nsnull;
}

void nsImageNanoX::SetAlphaLevel(PRInt32 aAlphaLevel)
{}



PRInt32 nsImageNanoX::GetAlphaLevel()
{
	return 0;
}

void nsImageNanoX::MoveAlphaMask(PRInt32 aX, PRInt32 aY)
{}


void nsImageNanoX::ImageUpdated(nsIDeviceContext *aContext, PRUint8 aFlags,
                                nsRect *aUpdateRect)
{
	unsigned bottom, left, right;
	bottom = aUpdateRect->y + aUpdateRect->height;
	left = aUpdateRect->x;
	right = left + aUpdateRect->width;

	// check if the image has an all-opaque 8-bit alpha mask
	if ((mAlphaDepth == 8) && !mAlphaValid) {
		bottom = aUpdateRect->y + aUpdateRect->height;
		left = aUpdateRect->x;
		right = left + aUpdateRect->width;
		for (unsigned y = aUpdateRect->y; (y < bottom) && !mAlphaValid; y++) {
			unsigned char *alpha = mAlphaBits + mAlphaRowBytes * y + left;
			for (unsigned x = left; x < right; x++) {
				if (*(alpha++) != 255) {
					mAlphaValid = PR_TRUE;
					break;
				}
			}
		}
	}

	// check if the image is a spacer
	if ((mAlphaDepth == 1) && mIsSpacer) {
		// mask of the leading/trailing bits in the update region
		PRUint8 leftmask = 0xff >> (left & 0x7);
		PRUint8 rightmask = 0xff << (7 - ((right - 1) & 0x7));

		// byte where the first/last bits of the update region are located
		PRUint32 leftindex = left >> 3;
		PRUint32 rightindex = (right - 1) >> 3;

		// first/last bits in the same byte - combine mask into leftmask
		// and fill rightmask so we don't try using it
		if (leftindex == rightindex) {
			leftmask &= rightmask;
			rightmask = 0xff;
		}

		// check the leading bits
		if (leftmask != 0xff) {
			PRUint8 *ptr = mAlphaBits + mAlphaRowBytes * aUpdateRect->y + leftindex;
			for (unsigned y = aUpdateRect->y; y < bottom; y++, ptr += mAlphaRowBytes) {
				if (*ptr & leftmask) {
					mIsSpacer = PR_FALSE;
					break;
				}
			}
			// move to first full byte
			leftindex++;
		}

		// check the trailing bits
		if (mIsSpacer && (rightmask != 0xff)) {
			PRUint8 *ptr = mAlphaBits + mAlphaRowBytes * aUpdateRect->y + rightindex;
			for (unsigned y = aUpdateRect->y; y < bottom; y++, ptr += mAlphaRowBytes) {
				if (*ptr & rightmask) {
					mIsSpacer = PR_FALSE;
					break;
				}
			}
			// move to last full byte
			rightindex--;
		}

		// check the middle bytes
		if (mIsSpacer && (leftindex <= rightindex)) {
			for (unsigned y = aUpdateRect->y; (y < bottom) && mIsSpacer; y++) {
				unsigned char *alpha = mAlphaBits + mAlphaRowBytes * y + leftindex;
				for (unsigned x = left; x < right; x++) {
					if (*(alpha++) != 0) {
						mIsSpacer = PR_FALSE;
						break;
					}
				}
			}
		}
	}


	if (mAlphaValid && mImagePixmap) {
		GrDestroyWindow(mImagePixmap);
		mImagePixmap = 0;
	}

	if (!mAlphaValid) {
		CreateOffscreenPixmap(mWidth, mHeight);
		DrawImageOffscreen(0, 0, mWidth, mHeight, NULL);

		NS_ASSERTION(mImagePixmap != 0, "Hmmm...\n");

		if (!mGC)
			mGC = GrNewGC();
/*
		GrArea(mImagePixmap, mGC,
		       aUpdateRect->x, aUpdateRect->y,
		       aUpdateRect->width, aUpdateRect->height,
		       mImageBits + mRowBytes * aUpdateRect->y + 3*aUpdateRect->x,
		       MWPF_TRUECOLOR888);
*/
	}

	mFlags = aFlags;
}

NS_IMETHODIMP
nsImageNanoX::DrawScaled(nsIRenderingContext &aContext,
                         nsDrawingSurface aSurface,
                         PRInt32 aSX, PRInt32 aSY,
                         PRInt32 aSWidth, PRInt32 aSHeight,
                         PRInt32 aDX, PRInt32 aDY,
                         PRInt32 aDWidth, PRInt32 aDHeight)
{
	PRInt32 origSHeight = aSHeight, origDHeight = aDHeight;
	PRInt32 origSWidth = aSWidth, origDWidth = aDWidth;

	if (aSWidth < 0 || aDWidth < 0 || aSHeight < 0 || aDHeight < 0) {
		NS_ASSERTION(!(aSWidth < 0 || aDWidth < 0 || aSHeight < 0 ||
		               aDHeight < 0), "Dimenson < 0");
		return NS_ERROR_FAILURE;
	}

	if (0 == aSWidth || 0 == aDWidth || 0 == aSHeight || 0 == aDHeight) {
		NS_ASSERTION(!(0 == aSWidth || 0 == aDWidth || 0 == aSHeight ||
		               0 == aDHeight), "Dimension == 0");
		return NS_OK;
	}

	// limit the size of the blit to the amount of the image read in
	if (aSX + aSWidth > mDecodedX2) {
		aDWidth -= ((aSX + aSWidth - mDecodedX2) * origDWidth) / origSWidth;
		aSWidth -= (aSX + aSWidth) - mDecodedX2;
	}
	if (aSX < mDecodedX1) {
		aDX += ((mDecodedX1 - aSX) * origDWidth) / origSWidth;
		aSX = mDecodedX1;
	}

	if (aSY + aSHeight > mDecodedY2) {
		aDHeight -= ((aSY + aSHeight - mDecodedY2) * origDHeight) / origSHeight;
		aSHeight -= (aSY + aSHeight) - mDecodedY2;
	}
	if (aSY < mDecodedY1) {
		aDY += ((mDecodedY1 - aSY) * origDHeight) / origSHeight;
		aSY = mDecodedY1;
	}

	if ((aDWidth <= 0 || aDHeight <= 0) || (aSWidth <= 0 || aSHeight <= 0))
		return NS_OK;

	nsDrawingSurfaceNanoX *drawing = (nsDrawingSurfaceNanoX*)aSurface;

/*
	if (mAlphaDepth == 1) {
		CreateAlphaBitmap(mWidth, mHeight);
	}
*/

	if ((mAlphaDepth == 8) 
		/*  && mAlphaValid*/ 	// XXX we should already have our mask
		 			// region... (but not always?)
		 ) {
#if defined(DEBUG_mattw)
	fprintf(stderr, "DrawScaled: 8bit alpha, calling DrawComposited\n");
#endif
		DrawComposited(aContext, aSurface,
		               aSX, aSY, aSWidth, aSHeight,
		              aDX, aDY, aDWidth, aDHeight);
		return NS_OK;
	}

	GR_GC_ID gc = GrNewGC();
	GR_REGION_ID scaledRegion = 0;
	
	// set up the bitmap mask 
	if(mAlphaDepth == 1) {
#ifdef DEBUG_mattw
		fprintf(stderr, "DrawScaled: using 1 bit alpha\n");
#endif
		PRUint32 scaledRowBytes = (aDWidth+7)>>3; // round to next byte
		PRUint8 *scaledAlpha = (PRUint8*)nsMemory::Alloc(aDHeight*scaledRowBytes);

		if(scaledAlpha) {

			memset(scaledAlpha, 0, aDHeight*scaledRowBytes);

			RectStretch(aSX, aSY, aSX + aSWidth - 1, aSY + aSHeight - 1, 
				    0, 0, aDWidth - 1, aDHeight - 1,
				    mAlphaBits, mAlphaRowBytes, 
				    scaledAlpha, scaledRowBytes, 1);

			// have to do a bit of a hack here, save mAlphaBits and
			// mAlphaRegion (if it exists) and create a scaled alpha region 
			{
				PRUint8 *saveAlphaBits = mAlphaBits;
				int saveAlphaRowBytes = mAlphaRowBytes;
				GR_REGION_ID saveAlphaRegion = mAlphaRegion;
				mAlphaRegion = 0;

				mAlphaBits = scaledAlpha;
				mAlphaRowBytes = scaledRowBytes;

				// do it.
				CreateAlphaBitmap(aDWidth, aDHeight);
				scaledRegion = mAlphaRegion;

				mAlphaBits = saveAlphaBits;
				mAlphaRegion = saveAlphaRegion;
				mAlphaRowBytes = saveAlphaRowBytes;
			}
			nsMemory::Free(scaledAlpha);

			GrOffsetRegion(scaledRegion, aDX - aSX, aDY - aSY);
			GrSetGCRegion(gc, scaledRegion);
		}
	} 
#ifdef DEBUG_mattw
	else fprintf(stderr, "DrawScaled: NOT using alpha\n");
#endif

	PRUint8 *scaledRGB = (PRUint8 *)nsMemory::Alloc(3 * aDWidth * aDHeight);
	if (scaledRGB && gc) {
		RectStretch(aSX, aSY, aSX + aSWidth - 1, aSY + aSHeight - 1,
		            0, 0, aDWidth - 1, aDHeight - 1,
		            mImageBits, mRowBytes, scaledRGB, 3*aDWidth, 24);
		GrArea(drawing->GetDrawable(), gc,
		       aDX, aDY, aDWidth, aDHeight,
		       scaledRGB,
		       MWPF_TRUECOLOR888);
		nsMemory::Free(scaledRGB);
	} 

	GrDestroyGC(gc);

	if(scaledRegion) {
		GrDestroyRegion(scaledRegion);
	}
		
	mFlags = 0;
	return NS_OK;
}

NS_IMETHODIMP nsImageNanoX::Draw(nsIRenderingContext &aContext, nsDrawingSurface aSurface,
                                 PRInt32 aSX, PRInt32 aSY, PRInt32 aSWidth, PRInt32 aSHeight,
                                 PRInt32 aDX, PRInt32 aDY, PRInt32 aDWidth, PRInt32 aDHeight)
{
	if (aSurface == nsnull) {
		NS_ASSERTION(aSurface != nsnull, "hmm..");
		return NS_ERROR_FAILURE;
	}

	if ((mAlphaDepth == 1) && mIsSpacer) {
		return NS_OK;
	}

	if (aSWidth != aDWidth || aSHeight != aDHeight) {
		return DrawScaled(aContext, aSurface, aSX, aSY, aSWidth, aSHeight,
		                  aDX, aDY, aDWidth, aDHeight);
	}

	if (aSWidth <= 0 || aDWidth <= 0 || aSHeight <= 0 || aDHeight <= 0) {
#ifdef DEBUG_mattw
		NS_ASSERTION(aSWidth > 0 && aDWidth > 0 && aSHeight > 0 && aDHeight > 0,
		             "You can't draw an image with a 0 width or height!");
#endif 
		return NS_OK;
	}

	// limit the size of the blit to the amount of the image read in
	PRInt32 j = aSX + aSWidth;
	PRInt32 z;
	if (j > mDecodedX2) {
		z = j - mDecodedX2;
		aDWidth -= z;
		aSWidth -= z;
	}
	if (aSX < mDecodedX1) {
		aDX += mDecodedX1 - aSX;
		aSX = mDecodedX1;
	}

	j = aSY + aSHeight;
	if (j > mDecodedY2) {
		z = j - mDecodedY2;
		aDHeight -= z;
		aSHeight -= z;
	}
	if (aSY < mDecodedY1) {
		aDY += mDecodedY1 - aSY;
		aSY = mDecodedY1;
	}

	if (aDWidth <= 0 || aDHeight <= 0 || aSWidth <= 0 || aSHeight <= 0) {
#ifdef DEBUG_mattw
		NS_ASSERTION(!(aDWidth <= 0 || aDHeight <= 0 || aSWidth <= 0 ||
		               aSHeight <= 0), "Dimenstion <= 0");
#endif 
		return NS_OK;
	}

	if ((mAlphaDepth == 8) && mAlphaValid) {
		DrawComposited(aContext, aSurface,
		               aSX, aSY, aSWidth, aSHeight,
		               aDX, aDY, aSWidth, aSHeight);
		return NS_OK;
	}

	nsDrawingSurfaceNanoX *drawing = (nsDrawingSurfaceNanoX*)aSurface;

	if (mAlphaDepth == 1)
		CreateAlphaBitmap(mWidth, mHeight);

	GR_GC_ID gc = GrNewGC();
	if (mAlphaRegion) {
		GrOffsetRegion(mAlphaRegion, aDX - aSX, aDY - aSY);
		GrSetGCRegion(gc, mAlphaRegion);
	}

#if defined(DEBUG_mattw) && defined(ANNOYING_PRINTFS)
	fprintf(stderr, "Draw: sx=%d, sy=%d, sw=%d, sh=%d\n",
	       aSX, aSY, aSWidth, aSHeight);
	fprintf(stderr, "Draw: dx=%d, dy=%d, dw=%d, dh=%d\n",
	       aDX, aDY, aDWidth, aDHeight);
#endif

	GrCopyArea(drawing->GetDrawable(),
	           gc,
	           aDX,                          // xdest
	           aDY,                          // ydest
	           aSWidth,                      // width
	           aSHeight,                    // height
	           mImagePixmap,                 // source window
	           aSX,                          // xsrc
	           aSY,                          // ysrc
	           MWROP_SRCCOPY);

	mFlags = 0;
	GrDestroyGC(gc);
	return NS_OK;
}

//------------------------------------------------------------
// 8-bit alpha composite drawing...

// Compositing code consists of these functions:
//  * nsImageNanoX::DrawComposited32() - 32-bit (888) truecolor convert/composite
//  * nsImageNanoX::DrawComposited24() - 24-bit (888) truecolor convert/composite
//  * nsImageNanoX::DrawComposited16() - 16-bit ([56][56][56]) truecolor
//                                       convert/composite
//  * nsImageNanoX::DrawCompositedGeneral() - convert/composite for any visual
//                                            not handled by the above methods
//  * nsImageNanoX::DrawComposited() - compositing master method; does region
//                                     clipping, calls one of the above, then
//                                     writes out the composited image

// 32-bit (888) truecolor convert/composite function
void
nsImageNanoX::DrawComposited32(PRUint8 *imageOrigin, PRUint32 imageStride,
                               PRUint8 *alphaOrigin, PRUint32 alphaStride,
                               unsigned width, unsigned height,
                               GR_IMAGE_HDR *ximage, unsigned char *targetPixels)
{

	for (unsigned y = 0; y < height; y++) {
		unsigned char *baseRow = (unsigned char *)ximage->imagebits + y * ximage->pitch;
		unsigned char *targetRow = targetPixels + gPixelSize * (y * ximage->width);
		unsigned char *imageRow = imageOrigin + y * imageStride;
		unsigned char *alphaRow = alphaOrigin + (y * alphaStride);

		for (unsigned i = 0; i < width;
		     i++, baseRow += gPixelSize, targetRow += gPixelSize, imageRow += 3, alphaRow++) {
			unsigned alpha = *alphaRow;
			MOZ_BLEND(targetRow[0], baseRow[2], imageRow[0], alpha);
			MOZ_BLEND(targetRow[1], baseRow[1], imageRow[1], alpha);
			MOZ_BLEND(targetRow[2], baseRow[0], imageRow[2], alpha);
		}
	}
}

// 24-bit (888) truecolor convert/composite function
void
nsImageNanoX::DrawComposited24(PRUint8 *imageOrigin, PRUint32 imageStride,
                               PRUint8 *alphaOrigin, PRUint32 alphaStride,
                               unsigned width, unsigned height,
                               GR_IMAGE_HDR *ximage, unsigned char *targetPixels)
{
	for (unsigned y = 0; y < height; y++) {
		unsigned char *baseRow = (unsigned char *)ximage->imagebits
		                         + y * ximage->pitch;
		unsigned char *targetRow = targetPixels + 4 * (y * ximage->width);
		unsigned char *imageRow = imageOrigin + y * imageStride;
		unsigned char *alphaRow = alphaOrigin + y * alphaStride;

		for (unsigned i = 0; i < width;
		     i++, baseRow += 4, targetRow += 4, imageRow += 3, alphaRow++) {
			unsigned alpha = *alphaRow;
			MOZ_BLEND(targetRow[0], baseRow[0], imageRow[0], alpha);
			MOZ_BLEND(targetRow[1], baseRow[1], imageRow[1], alpha);
			MOZ_BLEND(targetRow[2], baseRow[2], imageRow[2], alpha);
		}
	}
}

unsigned nsImageNanoX::scaled6[1<<6] = {
  3,   7,  11,  15,  19,  23,  27,  31,  35,  39,  43,  47,  51,  55,  59,  63,
 67,  71,  75,  79,  83,  87,  91,  95,  99, 103, 107, 111, 115, 119, 123, 127,
131, 135, 139, 143, 147, 151, 155, 159, 163, 167, 171, 175, 179, 183, 187, 191,
195, 199, 203, 207, 211, 215, 219, 223, 227, 231, 235, 239, 243, 247, 251, 255
};

unsigned nsImageNanoX::scaled5[1<<5] = {
  7,  15,  23,  31,  39,  47,  55,  63,  71,  79,  87,  95, 103, 111, 119, 127,
135, 143, 151, 159, 167, 175, 183, 191, 199, 207, 215, 223, 231, 239, 247, 255
};

// 16-bit ([56][56][56]) truecolor convert/composite function
void
nsImageNanoX::DrawComposited16(PRUint8 *imageOrigin, PRUint32 imageStride,
                               PRUint8 *alphaOrigin, PRUint32 alphaStride,
                               unsigned width, unsigned height,
                               GR_IMAGE_HDR *ximage, unsigned char *targetPixels)
{
	unsigned *redScale;
	unsigned *greenScale;
	unsigned *blueScale;
	unsigned redMask, greenMask, blueMask;
	unsigned redShift, greenShift, blueShift;

	switch(mPixType) {
		case MWPF_TRUECOLOR555: 	/* pixel is packed 16 bits 5/5/5 truecolor*/
#ifdef DEBUG_mattw
			fprintf(stderr, "It's 555!!\n");
#endif
			redScale = scaled5;
			greenScale = scaled5;
			blueScale = scaled5;
			
			redMask = 0x7C00;	// 111110000000000
			greenMask = 0x3E0;	// 000001111100000
			blueMask = 0x1F;	// 000000000011111

			redShift = 10; greenShift = 5; blueShift = 0;
			break;

		case MWPF_TRUECOLOR565: 	/* pixel is packed 16 bits 5/6/5 truecolor*/
#ifdef DEBUG_mattw
			fprintf(stderr, "It's 565!!\n");
#endif
			redScale = scaled5;
			greenScale = scaled6;
			blueScale = scaled5;
			
			redMask = 0xF800;	// 1111100000000000
			greenMask = 0x7E0;	// 0000011111100000
			blueMask = 0x1F;	// 0000000000011111

			redShift = 11; greenShift = 5; blueShift = 0;
			break;

		default:
			NS_ASSERTION(0,"hmmm.. unusual pixel type in DrawComposited16..");
			return;
			break;
	}

	for (unsigned y = 0; y < height; y++) {
		unsigned char *baseRow = (unsigned char *)ximage->imagebits + y * ximage->pitch;
		unsigned char *targetRow = targetPixels + 4 * (y * ximage->width);
		unsigned char *imageRow = imageOrigin + y * imageStride;
		unsigned char *alphaRow = alphaOrigin + (y * alphaStride);

		// assumptions: targetRow and baseRow format is 32bpp, imageRow
		// format is 24bpp.  baseRow format is 555/565 16bpp.
		// (let translation to actual format happen on the server)

		for (unsigned i = 0; i < width;
		     i++, baseRow += 4, targetRow += 4, imageRow += 3, alphaRow++) {
			unsigned long pix = *((unsigned long *)baseRow);
			unsigned alpha = *alphaRow;

			MOZ_BLEND(targetRow[0],
			          redScale[(pix&redMask) >> redShift],
			          imageRow[0], 
				  alpha);
			MOZ_BLEND(targetRow[1],
			          greenScale[(pix&greenMask) >> greenShift],
			          imageRow[1], 
				  alpha);
			MOZ_BLEND(targetRow[2],
			          blueScale[(pix&blueMask) >> blueShift],
			          imageRow[2], 
				  alpha);
		}
	}
}

// Generic convert/composite function
void
nsImageNanoX::DrawCompositedGeneral(PRUint8 *imageOrigin, PRUint32 imageStride,
                                    PRUint8 *alphaOrigin, PRUint32 alphaStride,
                                    unsigned width, unsigned height,
                                    GR_IMAGE_HDR *ximage, unsigned char *targetPixels)
{
	fprintf(stderr, "8bit alpha images at a screen depth of 8bpp is not implemented.\n");
}

void
nsImageNanoX::DrawComposited(nsIRenderingContext &aContext,
                             nsDrawingSurface aSurface,
                             PRInt32 aSX, PRInt32 aSY,
                             PRInt32 aSWidth, PRInt32 aSHeight,
                             PRInt32 aDX, PRInt32 aDY,
                             PRInt32 aDWidth, PRInt32 aDHeight)
{
	if ((aDWidth==0) || (aDHeight==0))
		return;

	// dimension checks taken from xlib code
	PRUint32 surfaceWidth, surfaceHeight;
	nsDrawingSurfaceNanoX *drawing = (nsDrawingSurfaceNanoX*)aSurface;
	drawing->GetDimensions(&surfaceWidth, &surfaceHeight);

	int readX, readY;
	unsigned int readWidth, readHeight, destX, destY;

	if ((aDY >= (int)surfaceHeight) || (aDX >= (int)surfaceWidth) ||
	    (aDY + aDHeight <= 0) || (aDX + aDWidth <= 0)) {
		// this should never happen if the layout engine is sane,
		// as it means we are trying to draw an image which is outside
		// the drawing surface.
		NS_ASSERTION(0, "Layout engine gone nuts.");
		return ;
	}

	if (aDX < 0) {
		readX = 0;
		readWidth = aDWidth + aDX;
		destX = aSX - aDX;
	} else {
		readX = aDX;
		readWidth = aDWidth;
		destX = aSX;
	}
	if (aDY < 0) {
		readY = 0;
		readHeight = aDHeight + aDY;
		destY = aSY - aDY;
	} else {
		readY = aDY;
		readHeight = aDHeight;
		destY = aSY;
	}

	if (readX + readWidth > surfaceWidth)
		readWidth = surfaceWidth - readX;
	if (readY + readHeight > surfaceHeight)
		readHeight = surfaceHeight - readY;

	if ((readHeight <= 0) || (readWidth <= 0)) {
		NS_ASSERTION(!((readHeight <= 0) || (readWidth <= 0)), "Width or height	should be > 0");
		return ;
	}

	// Read in area. 
	unsigned char
	*readPixels = (unsigned char *)nsMemory::Alloc(readWidth * readHeight * gPixelSize);
	unsigned char
	*targetPixels = (unsigned char *)nsMemory::Alloc(readWidth * readHeight * gPixelSize);

	GrReadArea(drawing->GetDrawable(), readX, readY, readWidth,
	           readHeight, (GR_PIXELVAL *)readPixels);

	PRUint8 *scaledImage = 0;
	PRUint8 *scaledAlpha = 0;
	PRUint8 *imageOrigin, *alphaOrigin;
	PRUint32 imageStride, alphaStride;
	if ((aSWidth != aDWidth) || (aSHeight != aDHeight)) {
		PRUint32 x1, Y1, x2, y2;
		x1 = (destX * aSWidth) / aDWidth;
		Y1 = (destY * aSHeight) / aDHeight;
		x2 = ((destX + readWidth) * aSWidth) / aDWidth;
		y2 = ((destY + readHeight) * aSHeight) / aDHeight;

		scaledImage = (PRUint8 *)nsMemory::Alloc(gPixelSize * aDWidth * aDHeight);
		scaledAlpha = (PRUint8 *)nsMemory::Alloc(aDWidth * aDHeight);
		if (!scaledImage || !scaledAlpha) {
			nsMemory::Free(targetPixels);
			nsMemory::Free(readPixels);
			if (scaledImage)
				nsMemory::Free(scaledImage);
			if (scaledAlpha)
				nsMemory::Free(scaledAlpha);
#ifdef DEBUG_mattw
			fprintf(stderr, "aiiiiiiiiii! returning from DrawComposited, no memory!\n");
#endif 
			return ;
		}

//		printf("DrawComposited: mRowBytes=%d\n", mRowBytes);
		RectStretch(x1, Y1, x2 - 1, y2 - 1,
		            0, 0, readWidth - 1, readHeight - 1,
		            mImageBits, mRowBytes, scaledImage, 3 * readWidth, 24);
		
		RectStretch(x1, Y1, x2 - 1, y2 - 1,
		            0, 0, readWidth - 1, readHeight - 1,
		            mAlphaBits, mAlphaRowBytes, scaledAlpha, readWidth, 8);

		imageOrigin = scaledImage;
		imageStride = 3 * readWidth;
		alphaOrigin = scaledAlpha;
		alphaStride = readWidth;
	} else {
		imageOrigin = mImageBits + destY * mRowBytes + 3 * destX;
		imageStride = mRowBytes;
		alphaOrigin = mAlphaBits + destY * mAlphaRowBytes + destX;
		alphaStride = mAlphaRowBytes;
	}
	GR_IMAGE_HDR hdr;
	hdr.pitch = readWidth * gPixelSize; // bytes per line
	hdr.width = readWidth;
	hdr.height = readHeight;
	hdr.imagebits = readPixels;

	switch (mPixType) {
		case MWPF_TRUECOLOR0888: 	/* pixel is packed 32 bits 8/8/8 truecolor*/
#ifdef DEBUG_mattw
			fprintf(stderr,"======== DrawComposited.. 32\n");
#endif
			DrawComposited32(imageOrigin, imageStride, alphaOrigin, alphaStride,
			                 readWidth, readHeight, &hdr, targetPixels);
			break;

		case MWPF_TRUECOLOR888: 	/* pixel is packed 24 bits 8/8/8 truecolor*/
#ifdef DEBUG_mattw
			fprintf(stderr,"======== DrawComposited.. 24\n");
#endif
			DrawComposited24(imageOrigin, imageStride, alphaOrigin, alphaStride,
			                 readWidth, readHeight, &hdr, targetPixels);
			break;

		case MWPF_TRUECOLOR555: 	/* pixel is packed 16 bits 5/5/5 truecolor*/
		case MWPF_TRUECOLOR565: 	/* pixel is packed 16 bits 5/6/5 truecolor*/
#ifdef DEBUG_mattw
			fprintf(stderr,"======== DrawComposited.. 16\n");
#endif
			DrawComposited16(imageOrigin, imageStride, alphaOrigin, alphaStride,
			                 readWidth, readHeight, &hdr, targetPixels);
			break;

			// case MWPF_TRUECOLOR332:	/* pixel is packed 8 bits 3/3/2 truecolor*/
		default:
#ifdef DEBUG_mattw
			fprintf(stderr, "======== DrawComposited.. General\n");
#endif
			DrawCompositedGeneral(imageOrigin, imageStride, alphaOrigin, alphaStride,
			                      readWidth, readHeight, &hdr, targetPixels);
			break;
	}

	GR_GC_ID gc = GrNewGC();
	GrArea(drawing->GetDrawable(), gc,
	       readX, readY, readWidth, readHeight, targetPixels,
	       MWPF_RGB);
	GrDestroyGC(gc);
	if(readPixels)
		nsMemory::Free(readPixels);
	if(targetPixels)
		nsMemory::Free(targetPixels);
	mFlags = 0;
}

NS_IMETHODIMP nsImageNanoX::Draw(nsIRenderingContext &aContext,
                                 nsDrawingSurface aSurface,
                                 PRInt32 aX, PRInt32 aY,
                                 PRInt32 aWidth, PRInt32 aHeight)
{
	if ((mAlphaDepth == 1) && mIsSpacer)
		return NS_OK;

	if (aSurface == nsnull) {
		NS_ASSERTION(aSurface != nsnull, "!");
		return NS_ERROR_FAILURE;
	}

	if ((mAlphaDepth == 8) && mAlphaValid) {
		DrawComposited(aContext, aSurface, 0, 0, aWidth, aHeight, aX, aY, aWidth, aHeight);
		return NS_OK;
	}


	// XXX kipp: this is temporary code until we eliminate the
	// width/height arguments from the draw method.
	if ((aWidth != mWidth) || (aHeight != mHeight)) {
		aWidth = mWidth;
		aHeight = mHeight;
	}

	PRInt32 validX = 0, 
		validY = 0, 
		validWidth = aWidth, 
		validHeight = aHeight;

	// limit the image rectangle to the size of the image data which
	// has been validated.
	if (mDecodedY2 < aHeight) {
		validHeight = mDecodedY2 - mDecodedY1;
	}
	if (mDecodedX2 < aWidth) {
		validWidth = mDecodedX2 - mDecodedX1;
	}
	if (mDecodedY1 > 0) {
		validHeight -= mDecodedY1;
		validY = mDecodedY1;
	}
	if (mDecodedX1 > 0) {
		validWidth -= mDecodedX1;
		validX = mDecodedX1;
	}

	nsDrawingSurfaceNanoX *drawing = (nsDrawingSurfaceNanoX*)aSurface;
	CreateAlphaBitmap(aWidth, aHeight);

	GR_GC_ID gc;
	gc = GrCopyGC(drawing->GetGC());

	if (mAlphaRegion) {
		GrOffsetRegion(mAlphaRegion, aX, aY);
		GrSetGCRegion(gc, mAlphaRegion);
	}

		if (aX < 0) {
			validWidth += aX;
			validX -= aX;
			aX = 0;
		}
		if (aY < 0) {
			validHeight += aY;
			validY -= aY;
			aY = 0;
		}
	GrCopyArea(drawing->GetDrawable(), gc, validX + aX, validY + aY,
	           validWidth, validHeight, mImagePixmap,
	           validX, validY, MWROP_SRCCOPY);

	GrDestroyGC(gc);

	mFlags = 0;
	return NS_OK;
}

void nsImageNanoX::CreateAlphaBitmap(PRInt32 aWidth, PRInt32 aHeight)
{
	unsigned int alphaByte;
	unsigned int alphaBitShift;
	PRBool bitSet = PR_TRUE;
	PRBool setState = PR_FALSE;
	GR_RECT rect;

	if (mAlphaBits != nsnull
		/* &&IsFlagSet(nsImageUpdateFlags_kBitsChanged, mFlags)*/) {
		if (mAlphaRegion)
			GrDestroyRegion(mAlphaRegion);
		mAlphaRegion = GrNewRegion();
		for (int yc = 0; yc < aHeight; yc++) {
			alphaBitShift = 7;
			alphaByte = 0;
			for (int xc = 0; xc < aWidth; xc++) {
				bitSet = (*(mAlphaBits + yc * mAlphaRowBytes + alphaByte))
				         >> alphaBitShift & 1;

				if (alphaBitShift) {
					alphaBitShift--;
				} else {
					alphaBitShift = 7;
					alphaByte++;
				}
				if (bitSet != setState) {
					if (bitSet) {
						rect.x = xc;
						rect.y = yc;
					} else {
						rect.width = xc - rect.x;
						rect.height = yc - rect.y + 1;
						GrUnionRectWithRegion(mAlphaRegion, &rect);
					}
					setState = bitSet;
				}
			}
			if (setState) {
				rect.width = aWidth - rect.x;
				rect.height = yc - rect.y + 1;
				GrUnionRectWithRegion(mAlphaRegion, &rect);
				setState = PR_FALSE;
			}
		}
	}
}

void nsImageNanoX::CreateOffscreenPixmap(PRInt32 aWidth, PRInt32 aHeight)
{
	if (mImagePixmap == nsnull) {
		mImagePixmap = GrNewPixmap(aWidth, aHeight, nsnull);
	}
}

void nsImageNanoX::DrawImageOffscreen(PRInt32 validX, PRInt32 validY,
                                      PRInt32 validWidth, PRInt32 validHeight,
                                      nsDrawingSurface aDrawing)
{
  unsigned char *currentPixel;
  unsigned long colorVal;
  //GR_PIXELVAL returnPixel;
  int r, g, b;
  GR_GC_ID gc = GrNewGC();
  switch (mNumBytesPixel){
    case 1:{
      unsigned char *pixelmap = (unsigned char*)nsMemory::Alloc(validWidth*validHeight);
      for (int lY = validY; lY<validHeight; lY++){
        for (int lX = validX; lX<validWidth; lX++){
          currentPixel = mImageBits + lY * mRowBytes + lX * 3;
          r = currentPixel[0];
          g = currentPixel[1];
          b = currentPixel[2];
          colorVal = (b  << 16) | (g << 8) | (r << 0);
          //GrFindColor(colorVal, &returnPixel);
          pixelmap[lX+lY*validWidth] = colorVal; //returnPixel;
        }  
      }
      GrArea(mImagePixmap, gc, validX, validY, validWidth,
                 validHeight, pixelmap, MWPF_TRUECOLOR332);
      nsMemory::Free(pixelmap);
      break;
    }      
    case 2:{
      int *pixelmap = (int *)nsMemory::Alloc(validWidth*
                                        validHeight*4);
      for (int lY = validY; lY<validHeight; lY++){
        for (int lX = validX; lX<validWidth; lX++){
          currentPixel = mImageBits + lY * mRowBytes + lX * 3;
          b = currentPixel[0];
          g = currentPixel[1];
          r = currentPixel[2];
          pixelmap[lX+lY*validWidth] = (r  << 16) | (g << 8) | (b << 0);
        }  
      }
      GrArea(mImagePixmap, gc, validX, validY, validWidth,
                 validHeight, pixelmap, MWPF_RGB);
      nsMemory::Free(pixelmap);
      break;
    }  
    default:{
      int *pixelmap = (int *)nsMemory::Alloc(validWidth*
                                        validHeight*4);
      for (int lY = validY; lY<validHeight; lY++){
        for (int lX = validX; lX<validWidth; lX++){
          currentPixel = mImageBits + lY * mRowBytes + lX * 3;
          b = currentPixel[0];
          g = currentPixel[1];
          r = currentPixel[2];
          pixelmap[lX+lY*validWidth] = (r  << 16) | (g << 8) | (b << 0);
        }  
      }
      GrArea(mImagePixmap, gc, validX, validY, validWidth,
                 validHeight, pixelmap, MWPF_RGB);
      nsMemory::Free(pixelmap);
      break;
    }  
  }
  GrDestroyGC(gc);
}

void nsImageNanoX::TilePixmap(GR_WINDOW_ID src, GR_WINDOW_ID dest,
                              PRInt32 aSXOffset, PRInt32 aSYOffset,
                              const nsRect &destRect, const nsRect &clipRect,
                              PRBool useClip)
{
	PRInt32	aX0 = destRect.x - aSXOffset, 
		aY0 = destRect.y - aSYOffset, 
		aX1 = destRect.x + destRect.width, 
		aY1 = destRect.y + destRect.height;
	PRInt32 width, height;
	PRInt32 sx, sy, dx, dy;
	GR_WINDOW_INFO wInfo;
	GR_GC_ID gc = GrNewGC();
	GR_REGION_ID reg = 0;
	GrGetWindowInfo(src, &wInfo);

	if (useClip) {
		GR_RECT r = {clipRect.x, clipRect.y, clipRect.width, clipRect.height};
		reg = GrNewRegion();
		GrUnionRectWithRegion(reg, &r);
		GrSetGCRegion(gc, reg);
	}

	for (PRInt32 y = aY0; y < aY1; y += wInfo.height) {
		height = MWMIN(wInfo.height, aY1 - y);
		for (PRInt32 x = aX0; x < aX1; x += wInfo.width) {
			width = MWMIN(wInfo.width, aX1 - x);
			dx = x;
			dy = y;
			sx = sy = 0;
			if (x < destRect.x) {
				dx = destRect.x;
				sx = aSXOffset;
			}
			if (y < destRect.y) {
				dy = destRect.y;
				sy = aSYOffset;
			}
			if ((width > 0) && (height > 0)) {
				GrCopyArea(dest, gc, dx, dy, width, height,
				           src, sx, sy, MWROP_SRCCOPY);
			}
		}
	}
	if (reg)
		GrDestroyRegion(reg);
	GrDestroyGC(gc);
}

void nsImageNanoX::TileAlphaBitmap(GR_REGION_ID src, GR_REGION_ID dest,
                                   PRInt32 aSXOffset, PRInt32 aSYOffset,
                                   const nsRect &srcRect,
                                   const nsRect &destRect,
                                   const nsRect &clipRect, PRBool useClip)
{
	PRInt32	aX0 = destRect.x - aSXOffset,
	        aY0 = destRect.y - aSYOffset,
	        aX1 = destRect.x + destRect.width,
	        aY1 = destRect.y + destRect.height;
	PRInt32 width, height;
	PRInt32 sx, sy, dx, dy;
	GR_GC_ID gc = GrNewGC();

	for (PRInt32 y = aY0; y < aY1; y += srcRect.height) {
		height = MWMIN(srcRect.height, aY1 - y);
		for (PRInt32 x = aX0; x < aX1; x += srcRect.width) {
			width = MWMIN(srcRect.width, aX1 - x);
			dx = x;
			dy = y;
			sx = sy = 0;
			if (x < destRect.x) {
				dx = destRect.x;
				sx = aSXOffset;
			}
			if (y < destRect.y) {
				dy = destRect.y;
				sy = aSYOffset;
			}
			if ((width > 0) && (height > 0)) {
				// XXX shouldnt we set the GC alpha Region here?
				GrSetGCRegion(gc, mAlphaRegion);

				//#if 0
				GrCopyArea(dest, gc, dx, dy, width, height,
				           src, sx, sy, MWROP_SRCCOPY);
				//#endif
				GrOffsetRegion(src, dx, dy);
				GrUnionRegion(dest, dest, src);
				GrOffsetRegion(src, -dx, -dy);
			}
		}
	}
	GrDestroyGC(gc);
}


NS_IMETHODIMP nsImageNanoX::DrawTile(nsIRenderingContext &aContext,
                                     nsDrawingSurface aSurface,
                                     nsRect &aSrcRect,
                                     nsRect &aTileRect)
{
	if ((mAlphaDepth == 1) && mIsSpacer)
		return NS_OK;

	nsDrawingSurfaceNanoX *drawing = (nsDrawingSurfaceNanoX*)aSurface;
	PRBool partial = PR_FALSE;

	PRInt32	validX = 0, 
		validY = 0, 
		validWidth = mWidth, 
		validHeight = mHeight;

	// limit the image rectangle to the size of the image data which
	// has been validated.
	if (mDecodedY2 < mHeight) {
		validHeight = mDecodedY2 - mDecodedY1;
		partial = PR_TRUE;
	}
	if (mDecodedX2 < mWidth) {
		validWidth = mDecodedX2 - mDecodedX1;
		partial = PR_TRUE;
	}
	if (mDecodedY1 > 0) {
		validHeight -= mDecodedY1;
		validY = mDecodedY1;
		partial = PR_TRUE;
	}
	if (mDecodedX1 > 0) {
		validWidth -= mDecodedX1;
		validX = mDecodedX1;
		partial = PR_TRUE;
	}

	if (partial ||
	    (drawing->GetDepth() == 8) ||
	    ((mAlphaDepth == 8) && mAlphaValid)) {
#ifdef DEBUG_mattw
		printf("Warning: using slow tiling\n");
#endif
		PRInt32 aY0 = aTileRect.y, 
			aX0 = aTileRect.x, 
			aY1 = aTileRect.y + aTileRect.height, 
			aX1 = aTileRect.x + aTileRect.width;

		for (PRInt32 y = aY0; y < aY1; y += aSrcRect.height)
			for (PRInt32 x = aX0; x < aX1; x += aSrcRect.width)
				Draw(aContext, aSurface, x, y,
				     PR_MIN(aSrcRect.width, aX1 - x),
				     PR_MIN(aSrcRect.height, aY1 - y));

		return NS_OK;
	}

	if (mAlphaDepth == 1) {

		GR_DRAW_ID tileImg;
		GR_DRAW_ID tileMask;

		CreateAlphaBitmap(validWidth, validHeight);

		nsRect tmpRect(0, 0, aTileRect.width, aTileRect.height);

		tileImg = GrNewPixmap(aTileRect.width, aTileRect.height, 0);
		TilePixmap(mImagePixmap, tileImg, 0, 0, tmpRect, tmpRect, PR_FALSE);


		// tile alpha mask
		tileMask = GrNewPixmap(aTileRect.width, aTileRect.height, 0);
		TilePixmap(mImagePixmap, tileMask, 0, 0, tmpRect, tmpRect, PR_FALSE);

		if (!mAlphaRegion)
			CreateAlphaBitmap(aTileRect.width, aTileRect.height);

		GR_GC_ID fgc = GrCopyGC(drawing->GetGC());
		GrOffsetRegion(mAlphaRegion, aTileRect.x, aTileRect.y);
		GrSetGCRegion(fgc, mAlphaRegion);
		//gdk_gc_set_clip_mask(fgc, (GdkBitmap*)tileMask);
		//gdk_gc_set_clip_origin(fgc, aTileRect.x, aTileRect.y);

		// and copy it back
		GrCopyArea(drawing->GetDrawable(), fgc, aTileRect.x,
		           aTileRect.y,
		           aTileRect.width,
		           aTileRect.height,
		           tileImg, 0, 0, MWROP_SRCCOPY);
		GrDestroyGC(fgc);

		GrDestroyWindow(tileImg);
		GrDestroyWindow(tileMask);

	} else {

		// In the non-alpha case, gdk can tile for us

		nsRect clipRect;
		PRBool isValid;
		aContext.GetClipRect(clipRect, isValid);

		TilePixmap(mImagePixmap, drawing->GetDrawable(), aTileRect.x, aTileRect.y, aTileRect, clipRect, PR_TRUE);
	}

	return NS_OK;
}

NS_IMETHODIMP nsImageNanoX::DrawTile(nsIRenderingContext &aContext,
                                     nsDrawingSurface aSurface,
                                     PRInt32 aSXOffset, PRInt32 aSYOffset,
                                     const nsRect &aTileRect)
{
	if ((mAlphaDepth == 1) && mIsSpacer)
		return NS_OK;

	nsDrawingSurfaceNanoX *drawing = (nsDrawingSurfaceNanoX*)aSurface;
	PRBool partial = PR_FALSE;

	PRInt32	validX = 0, 
		validY = 0, 
		validWidth = mWidth, 
		validHeight = mHeight;


	// limit the image rectangle to the size of the image data which
	// has been validated.
	if (mDecodedY2 < mHeight) {
		validHeight = mDecodedY2 - mDecodedY1;
		partial = PR_TRUE;
	}
	if (mDecodedX2 < mWidth) {
		validWidth = mDecodedX2 - mDecodedX1;
		partial = PR_TRUE;
	}
	if (mDecodedY1 > 0) {
		validHeight -= mDecodedY1;
		validY = mDecodedY1;
		partial = PR_TRUE;
	}
	if (mDecodedX1 > 0) {
		validWidth -= mDecodedX1;
		validX = mDecodedX1;
		partial = PR_TRUE;
	}


	if ((validWidth == 0) || (validHeight == 0)) {
#ifdef DEBUG_matte
		NS_ASSERTION(!((validWidth == 0) || (validHeight == 0)), "Dimension == 0");
#endif 
		return NS_OK;
	}

	if (partial || ((mAlphaDepth == 8) && mAlphaValid)) {
#ifdef DEBUG_mattw
		printf("Warning: using slow tiling\n");
#endif
		// use slow tiling
		PRInt32 aX0 = aTileRect.x - aSXOffset, 
			aY0 = aTileRect.y - aSYOffset, 
			aX1 = aTileRect.x + aTileRect.width, 
			aY1 = aTileRect.y + aTileRect.height;

		PRBool clipState;
		aContext.PushState();
		aContext.SetClipRect(aTileRect, nsClipCombine_kIntersect,
		                     clipState);

		for (PRInt32 y = aY0; y < aY1; y += mHeight)
			for (PRInt32 x = aX0; x < aX1; x += mWidth) {

				Draw(aContext, aSurface,
				     x, y,
				     PR_MIN(validWidth, aX1 - x),
				     PR_MIN(validHeight, aY1 - y));
			}
#ifdef DEBUG_mattw
		printf("Done slow drawing.\n");
#endif
		aContext.PopState(clipState);
		return NS_OK;
	}


	if (mAlphaDepth == 1) {
		GR_WINDOW_ID tileImg;
		GR_REGION_ID tileRegion;

		CreateAlphaBitmap(validWidth, validHeight);

		nsRect tmpRect(0, 0, aTileRect.width, aTileRect.height);

		tileImg = GrNewPixmap(aTileRect.width, aTileRect.height, nsnull);
		TilePixmap(mImagePixmap, tileImg, aSXOffset, aSYOffset, tmpRect,
		           tmpRect, PR_FALSE);

		// tile alpha mask
		GR_GC_ID gc = GrCopyGC(drawing->GetGC());
		tileRegion = GrNewRegion();
		if (mAlphaRegion) {
			nsRect aRect (0, 0, validWidth, validHeight);
			TileAlphaBitmap(mAlphaRegion, tileRegion, aSXOffset, aSYOffset,
			                aRect, tmpRect, tmpRect, PR_FALSE);

			GrOffsetRegion(tileRegion, aTileRect.x - aSXOffset, aTileRect.y - aSYOffset);

			GrSetGCRegion(gc, tileRegion);
		}

		// and copy it back
		GrCopyArea(drawing->GetDrawable(), gc,
		           aTileRect.x, aTileRect.y, aTileRect.width,
		           aTileRect.height, tileImg,
		           0, 0, MWROP_SRCCOPY);
		GrDestroyWindow(tileImg);
		GrDestroyRegion(tileRegion);
		GrDestroyGC(gc);
	} else {
		nsRect clipRect;
		PRBool isValid;
		aContext.GetClipRect(clipRect, isValid);
		TilePixmap(mImagePixmap, drawing->GetDrawable(), aSXOffset, aSYOffset,
		           aTileRect, clipRect, PR_TRUE);
	}
	mFlags = 0;
	return NS_OK;
}

NS_IMETHODIMP nsImageNanoX::Optimize(nsIDeviceContext *aContext)
{
	return NS_OK;
}

NS_IMETHODIMP nsImageNanoX::LockImagePixels(PRBool aMaskPixels)
{
	return NS_OK;
}

NS_IMETHODIMP nsImageNanoX::UnlockImagePixels(PRBool aMaskPixels)
{
	return NS_OK;
}

NS_IMETHODIMP nsImageNanoX::SetDecodedRect(PRInt32 X1, PRInt32 Y1,
        PRInt32 X2, PRInt32 Y2)
{
	mDecodedX1 = X1;
	mDecodedY1 = Y1;
	mDecodedX2 = X2;
	mDecodedY2 = Y2;
	return NS_OK;
}

#ifdef USE_IMG2
NS_IMETHODIMP nsImageNanoX::DrawToImage(nsIImage* aDstImage, nscoord aDX, nscoord aDY,
                                        nscoord aDWidth, nscoord aDHeight)
{
	nsImageNanoX *dest = NS_STATIC_CAST(nsImageNanoX*, aDstImage);

	if (!dest) {
		NS_ASSERTION(dest != nsnull, "nsIImage *aDstImage is NULL!\n");
		return (NS_ERROR_FAILURE);
	}

	if (!dest->mImagePixmap)
		dest->CreateOffscreenPixmap(dest->mWidth, dest->mHeight);

	if (!dest->mImagePixmap) {
		NS_ASSERTION(dest->mImagePixmap != nsnull, "dest's mImagePixmap	uninitialised\n");
		return (NS_ERROR_FAILURE);
	}

	if (!mImagePixmap) {
		NS_ASSERTION(mImagePixmap != nsnull, "uninitialised mImagePixmap\n");
		return (NS_ERROR_FAILURE);
	}

	if (mAlphaDepth == 1)
		CreateAlphaBitmap(mWidth, mHeight);


	GR_GC_ID gc = GrNewGC();

	if (mAlphaRegion)
		GrSetGCRegion(gc, mAlphaRegion);

	GrCopyArea(dest->mImagePixmap, gc,
	           aDX, aDY,
	           mWidth, mHeight,
	           mImagePixmap,
	           0, 0, MWROP_SRCCOPY);

	//	printf("CopyingArea in DrawToImage: x=%d, y=%d, cx=%d, cy=%d, srcx=%d, srcy=%d\n",
	//	       aDX, aDY, aDWidth, aDHeight, 0, 0);

	GrDestroyGC(gc);

	if (!mIsSpacer || !mAlphaDepth)
		dest->mIsSpacer = PR_FALSE;

	// need to copy the mImageBits in case we're rendered scaled
	PRUint8 *scaledImage = 0, *scaledAlpha = 0;
	PRUint8 *rgbPtr = 0, *alphaPtr = 0;
	PRUint32 rgbStride, alphaStride;

	if ((aDWidth != mWidth) || (aDHeight != mHeight)) {
		// scale factor in DrawTo... start scaling
		scaledImage = (PRUint8 *)nsMemory::Alloc(3 * aDWidth * aDHeight);
		if (!scaledImage)
			return NS_ERROR_OUT_OF_MEMORY;

		RectStretch(0, 0, mWidth - 1, mHeight - 1, 0, 0, aDWidth - 1, aDHeight - 1,
		            mImageBits, mRowBytes, scaledImage, 3*aDWidth, 24);

		if (mAlphaDepth) {
			if (mAlphaDepth == 1)
				alphaStride = (aDWidth + 7) >> 3;      // round to next byte
			else
				alphaStride = aDWidth;

			scaledAlpha = (PRUint8 *)nsMemory::Alloc(alphaStride * aDHeight);
			if (!scaledAlpha) {
				nsMemory::Free(scaledImage);
				return NS_ERROR_OUT_OF_MEMORY;
			}

			RectStretch(0, 0, mWidth - 1, mHeight - 1, 0, 0, aDWidth - 1, aDHeight - 1,
			            mAlphaBits, mAlphaRowBytes, scaledAlpha, alphaStride,
			            mAlphaDepth);
		}
		rgbPtr = scaledImage;
		rgbStride = 3 * aDWidth;
		alphaPtr = scaledAlpha;
	} else {
		rgbPtr = mImageBits;
		rgbStride = mRowBytes;
		alphaPtr = mAlphaBits;
		alphaStride = mAlphaRowBytes;
	}

	PRInt32 y;
	// now composite the two images together
	switch (mAlphaDepth) {
		case 1:
			for (y = 0; y < aDHeight; y++) {
				PRUint8 *dst = dest->mImageBits + (y + aDY) * dest->mRowBytes + 3 * aDX;
				PRUint8 *dstAlpha = dest->mAlphaBits + (y + aDY) * dest->mAlphaRowBytes;
				PRUint8 *src = rgbPtr + y * rgbStride;
				PRUint8 *alpha = alphaPtr + y * alphaStride;
				for (int x = 0; x < aDWidth; x++, dst += 3, src += 3) {
#define NS_GET_BIT(rowptr, x) (rowptr[(x)>>3] &  (1<<(7-(x)&0x7)))
#define NS_SET_BIT(rowptr, x) (rowptr[(x)>>3] |= (1<<(7-(x)&0x7)))

					// if this pixel is opaque then copy into the destination image
					if (NS_GET_BIT(alpha, x)) {
						dst[0] = src[0];
						dst[1] = src[1];
						dst[2] = src[2];
						NS_SET_BIT(dstAlpha, aDX + x);
					}

#undef NS_GET_BIT
#undef NS_SET_BIT
				}
			}
			break;
		case 8:
			for (y = 0; y < aDHeight; y++) {
				PRUint8 *dst = dest->mImageBits + (y + aDY) * dest->mRowBytes + 3 * aDX;
				PRUint8 *dstAlpha =
				    dest->mAlphaBits + (y + aDY) * dest->mAlphaRowBytes + aDX;
				PRUint8 *src = rgbPtr + y * rgbStride;
				PRUint8 *alpha = alphaPtr + y * alphaStride;
				for (int x = 0; x < aDWidth; x++, dst += 3, dstAlpha++, src += 3, alpha++) {

					// blend this pixel over the destination image
					unsigned val = *alpha;
					MOZ_BLEND(dst[0], dst[0], src[0], val);
					MOZ_BLEND(dst[1], dst[1], src[1], val);
					MOZ_BLEND(dst[2], dst[2], src[2], val);
					MOZ_BLEND(*dstAlpha, *dstAlpha, val, val);
				}
			}
			break;
		case 0:
		default:
			for (y = 0; y < aDHeight; y++)
				memcpy(dest->mImageBits + (y + aDY)*dest->mRowBytes + 3*aDX,
				       rgbPtr + y*rgbStride,
				       3*aDWidth);
	}

	nsRect rect(aDX, aDY, aDWidth, aDHeight);
	dest->ImageUpdated(0, nsImageUpdateFlags_kBitsChanged, &rect);

	if (scaledAlpha)
		nsMemory::Free(scaledAlpha);
	if (scaledImage)
		nsMemory::Free(scaledImage);
	return NS_OK;
}
#endif

