/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Netscape Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s): 
 */

#include "nsKeyCode.h"

#include <nano-X.h>

#include "nsGUIEvent.h" // For NS_VK

struct nsKeyConverter 
{
  int          vkCode; // Platform independent key code
  unsigned int keysym; // NanoX keystroke event ch value
  unsigned int special; // is this a special key
};

struct nsKeyConverter nsKeycodes[] =
{
  { 0,                   0x21, 0}, /* ! */
  { 0,                   0x23, 0}, /* # */
  { 0,                   0x24, 0}, /* $ */
  { 0,                   0x25, 0}, /* % */
  { 0,                   0x26, 0}, /* & */
  { 0,                   0x27, 0}, /* ' */
  { NS_VK_BACK,          0x7f, 0},
  { NS_VK_TAB,           0x09, 0},
  { NS_VK_ENTER,	 MWKEY_ENTER, 1},
  { NS_VK_ESCAPE,        0x1b, 0},
  { NS_VK_SPACE,         0x20, 0},
  { NS_VK_PRINTSCREEN,   0x0d, 1},
  { NS_VK_SEMICOLON,     0x3b, 0},
  { NS_VK_EQUALS,        0x3d, 0},
  { NS_VK_COMMA,         0x2c, 0},
  { NS_VK_PERIOD,        0x2e, 0},
  { NS_VK_SLASH,         0x2f, 0},
  { NS_VK_BACK_QUOTE,    0x60, 0},
  { NS_VK_BACK_SLASH,    0x6c, 0},
  { NS_VK_QUOTE,         0x22, 0},
  { NS_VK_OPEN_BRACKET,  0x28, 0},
  { NS_VK_CLOSE_BRACKET, 0x29, 0},
  { NS_VK_PAGE_UP,       MWKEY_PAGEUP, 1},
  { NS_VK_PAGE_DOWN,     MWKEY_PAGEDOWN, 1},
  { NS_VK_END,           MWKEY_END, 1},
  { NS_VK_HOME,          MWKEY_HOME, 1},
  { NS_VK_LEFT,          MWKEY_LEFT, 1},
  { NS_VK_UP,            MWKEY_UP, 1},
  { NS_VK_RIGHT,         MWKEY_RIGHT, 1},
  { NS_VK_DOWN,          MWKEY_DOWN, 1},
  { NS_VK_INSERT,        MWKEY_INSERT, 1},
  { NS_VK_DELETE,        MWKEY_DELETE, 1},
  { NS_VK_F1,            MWKEY_F1, 1},
  { NS_VK_F2,            MWKEY_F2, 1},
  { NS_VK_F3,            MWKEY_F3, 1},
  { NS_VK_F4,            MWKEY_F4, 1},
  { NS_VK_F5,            MWKEY_F5, 1},
  { NS_VK_F6,            MWKEY_F6, 1},
  { NS_VK_F7,            MWKEY_F7, 1},
  { NS_VK_F8,            MWKEY_F8, 1},
  { NS_VK_F9,            MWKEY_F9, 1},
  { NS_VK_F10,           MWKEY_F10, 1},
  { NS_VK_F11,           MWKEY_F11, 1},
  { NS_VK_F12,           MWKEY_F12, 1}
};

PRInt32
nsKeyCode::ConvertKeyToVirtualKey(GR_EVENT *event)
{
  unsigned int i;
  unsigned int length = sizeof(nsKeycodes) / sizeof(struct nsKeyConverter);
  unsigned int special;

  special = ((event->keystroke.ch & 0xff00) != 0);

  if (!special) {
    if (event->keystroke.ch >=  1  && event->keystroke.ch <= 26 && 
       (event->keystroke.modifiers & MWKMOD_CTRL))
      return event->keystroke.ch - 1 + NS_VK_A;
    if (event->keystroke.ch >= 'a' && event->keystroke.ch <= 'z')
      return event->keystroke.ch - 'a' + NS_VK_A;
    if (event->keystroke.ch >= 'A' && event->keystroke.ch <= 'Z')
      return event->keystroke.ch - 'A' + NS_VK_A;
  }

  for (i = 0; i<length; i++) {
    if ((nsKeycodes[i].keysym == event->keystroke.ch) &&
        (special == nsKeycodes[i].special))
      return (nsKeycodes[i].vkCode);
  }
  return (int)event->keystroke.ch;
}

//////////////////////////////////////////////////////////////////////////
/* static */ PRBool
nsKeyCode::KeyCodeIsModifier(GR_EVENT *event)
{
  if ((event->keystroke.modifiers & MWKMOD_SHIFT) ||
  	  (event->keystroke.modifiers & MWKMOD_CTRL) ||
  	  (event->keystroke.modifiers & MWKMOD_ALT))
  {
	return PR_TRUE;
  }

  return PR_FALSE;
}
//////////////////////////////////////////////////////////////////////////
