#include "nscore.h"
#include "nsList.h"

nsListItem::nsListItem(void *aData, nsListItem *aPrev)
{
  next = nsnull;
  prev = aPrev;
  data = aData;
}

nsList::nsList()
{
  head = nsnull;
  tail = nsnull;
}

nsList::~nsList()
{
  reset();
}

void nsList::add(void *aData)
{
  if (head == nsnull) {		// We have an empty list, create the head
    head = new nsListItem(aData, nsnull);
    tail = head;
  } else {			// Append to the end of the list
    tail->setNext(new nsListItem(aData, tail));
    tail = tail->getNext();	// Reset the tail
    tail->setNext(nsnull);
  }
}

void nsList::remove(void *aData)
{
  if (head == nsnull) {		// Removing from an null list
    return;
  } else {			// Find the data
    nsListItem *aItem = head;
    while ((aItem != nsnull) && (aItem->getData() != aData)) {
      aItem = aItem->getNext();
    }
    if (aItem == nsnull) {	// We didn't find it
      return;
    } else if (aItem == head) {	// We have to remove the head
      head = aItem->getNext();
      delete aItem;
      if (head == nsnull)	// We have empties the list
        tail = nsnull;
      else
        head->setPrev(nsnull);
    } else if (aItem == tail) {	// We have to remove the tail
      tail = aItem->getPrev();
      delete aItem;
      if (tail == nsnull)	// We have emptied the list
        head = nsnull;
      else
        tail->setNext(nsnull);
    } else {			// We remove from the middle
      nsListItem *prev = aItem->getPrev();
      nsListItem *next = aItem->getNext();
      delete aItem;
      prev->setNext(next);
      next->setPrev(prev);
    }
  }
}

void nsList::reset()
{
  while (head != nsnull) {
    void *aData = head->getData();
    remove(aData);
  }
}

